'Filename   7SegmentCounter_100.bas
'Version    1.00
'Date       2013-09-18
'Purpose    Frequency counter using six 7 segment LED displays (common anode)
'Author     Georg Latzel, DL6GL
'Copyright  (c) Georg Latzel, free for private use under GNU license
'Hardware   ATmega8, six 7 segment display (common anode)
'           ULN2803, 74HC93, 74HC14
'Compiler   BASCOM-AVR 2.0.7.6
'           For readable "camel case" variables mark "Don't change case"
'           in BASCOM Options-Programmer-Environment-Editor
'           Fuse bits can be set with AVR Studio 4.0 & ISP-PROG via USB
'           Flash used: 1,630 bytes (19%)
'Edit history
'V1.00      2013-09-18
'           first release

'==============================================================================
'7 segments a...g & DP and ISP
'Segment a  PD7
'Segment b  PD6
'Segment c  PB5   ISP SCK
'Segment d  PB2
'Segment e  PB3   ISP MOSI
'Segment f  PB0
'Segment g  PB1
'Segment DP PB4   ISP MISO
'           PC6   ISP Reset
DDRB = &B00111111                                 'Port B0...B5 = output (LED segments)
DDRD = &B11000000                                 'Port D6, D7 = output (LED segments)
PortD = &B00011111                                'Pullup Port D0...D4 (not used)
Seg_a Alias PortD.7
Seg_b Alias PortD.6
Seg_c Alias PortB.5
Seg_d Alias PortB.2
Seg_e Alias PortB.3
Seg_f Alias PortB.0
Seg_g Alias PortB.1
Seg_DP Alias PortB.4
'To lit a 7 segment LED, the input of the ULN2803 must be logical 1
'and the AVR port at the display anode must be logical 0 (PNP switch)
'  --------
'  |  a   |
'  |f   b |
'  |  g   |
'  |e   c |
'  |  d   |
'  --------
'        ---- segments -----
'Number  g  f  e  d  c  b  a
'0       0  1  1  1  1  1  1
'1       0  0  0  0  1  1  0
'2       1  0  1  1  0  1  1
'3       1  0  0  1  1  1  1
'4       1  1  0  0  1  1  0
'5       1  1  0  1  1  0  1
'6       1  1  1  1  1  0  1
'7       0  0  0  0  1  1  1
'8       1  1  1  1  1  1  1
'9       1  1  0  1  1  1  1

'7 segment anodes
'A1         PC0
'A2         PC1
'A3         PC2
'A4         PC3
'A5         PC4
'A6         PC5
DDRC = &B00111111                                 'Port C0...C5 = output (LED anodes)
Anode_1 Alias PortC.0
Anode_2 Alias PortC.1
Anode_3 Alias PortC.2
Anode_4 Alias PortC.3
Anode_5 Alias PortC.4
Anode_6 Alias PortC.5

'Counter input
'T1         PD5

'==============================================================================
'FUSE-Bits (AVR Studio 4.0-Settings, [ ]=no check, [x]=check)
'In ATMEGA-Data sheets: Bit = 1 -> not set
'                       Bit = 0 -> set
'RSTDISBL   [ ]
'WTDON      [ ]
'SPIEN      [x]
'EESAVE     [ ]
'BOOTSZ     Boot Flash size=128 words (minimal boot size)
'BOOTRST    [ ]
'CKOPT      [ ]
'BODLEVEL   Brown-out detection at VCC=2.7V
'BODDEN     [ ]
'SUT_CKSEL  Ext. Crystal Osc. High Freq., Start-up time 16k + 64ms

$regfile = "m8def.dat"                            'ATMega8
$crystal = 16000000                               'Crystal 16,0 MHz
$hwstack = 32
$swstack = 32
$framesize = 32

'Variables Counter ------------------------------------------------------------
Dim dwdCount As DWord                             'Frequency counts
Dim wrdCountLo As Word At dwdCount Overlay        'Counts low word
Dim wrdCountHi As Word At dwdCount + 2 Overlay    'Counts high word
Dim wrdTimer0_runs As Word                        'Timer0-Overflow cycles for Counter
Dim dwdCountSave As Dword

'Common Variables, multiply used ----------------------------------------------
Dim bytTmp0 As Byte
Dim bytTmp1 As Byte

'Variables 7 segment display --------------------------------------------------
Dim strDis As String * 6                          '6 digits to display
Dim bytDis(6) As Byte At strDis Overlay           'single characters in strDis
Dim bytDigit As Byte                              'Display digit 1...6
Dim bytNumber As Byte                             'ASCII Number to be displayed
Dim bytMux As Byte                                'Display multiplexer counter
Dim bitShow As Bit                                '=1: show display digit
Const bytNumDigits = 6                            'Total digits

'Configure Timer0-Interrupt for counter gate (Timer0 = 8bit) ------------------
'Overflow time: Overflow-Counts * Prescale / crystal frequency
'= 255 * 64 / 16.000.000 = 1. 2 msec = ~ 980 Hz
'Possible Prescales: 8, 64, 256, 1024, here 256
'Chosen prescale: 64 for sufficient display multiplex frequency
'Gate time for Counter:
'Effective Timer0 Counts * 64 Timer0-Steps * 800 Timer0-overflows / Crystal freq.
'Shortening Timer0 gate time with PresetTimer0 will give the appropiate gate time
'Effective Timer0 Counts: 256 (8bit Timer0-Counts) - 56 (PresetTimer0) = 200
'Gives timer0 gate time
'= 200 * 64 * 800 /  16.000.000 Hz = 0,64 sec
'Gives frequency in units of 100 Hz

Config Timer0 = Timer , Prescale = 64
Const Presettimer0 = 56                           'Initialize Timer0 (shorten time, >0)
Const wrdTimer0_MaxRuns = 800                     'Max. Timer0 overflows
On Timer0 Timer0_isr                              'On Overflow go to Timer0_isr

'Configure Timer1-Interrupt for Counter (Timer1 = 16bit) ----------------------
Config Timer1 = Counter , Edge = Rising , Noise_Cancel = 1
On Timer1 Timer1_isr                              'On Overflow go to Timer1_isr

'Subs and functions -----------------------------------------------------------
Declare Sub ShowDigit
Declare Sub ClearDigit
Declare Sub ShowFrequency
Declare Sub SelfTest

Call SelfTest

'Enable interrupts ------------------------------------------------------------
Enable Timer0
Enable TIMER1
Timer0 = Presettimer0                             'initialize Timer0
wrdTimer0_runs = 1                                'Timer0 overflow run counter
bytMux = 0
bitShow = 0
bytDigit = 0
SREG.7 = 1                                        'Enable all interrupts
Start Timer0
Start TIMER1

Do

   If bitShow = 1 Then
      Call ShowFrequency                          'Display current multiplexed digit
      bitShow = 0
   End If

Loop


'==============================================================================
Timer0_isr:
'Interrupt Service Routine for Timer0 (Counter gate) and display MUX
'Set Timer0 preset time
'Increment timer0 runs until > max runs
'Stop Timer1 if max Timer0 runs are expired
'Get number of counts in gate Timer/Counter1
'Input:  wrdTimer0_runs current timer0 overflow count for timer1 gate
'        bytMux         current timer0 overflow count for display multiplexing
'Output: dwdCountSave   frequency in units 100 Hz
'        bitShow        =1 display multiplex time expired, show current display

   Timer0 = Presettimer0
   'Counter -------------------------------------------------------------------
   If wrdTimer0_runs = 0 Then                     'start counter
      Start Timer1
      wrdTimer0_runs = 1
   Elseif wrdTimer0_runs < wrdTimer0_MaxRuns Then 'wait for Timer0 max overflows
      Incr wrdTimer0_runs                         'increment Timer0 Overflow counter
   Else                                           'close gate, get Timer1 counter
      Stop Timer1
      wrdCountLo = Timer1                         'get low word (overlay dwdCount)
      dwdCountSave = dwdCount                     'Save frequency
      Shift dwdCountSave , Right , 3              'Divide by 8 (2^3)=frequency
      Timer1 = 0                                  'reset timer1
      wrdTimer0_runs = 0
      wrdCountHi = 0                              'reset counter
      wrdCountLo = 0
   End If

   'Display multiplexer -------------------------------------------------------
   If bytMux < 4 Then                             'gives flickerfree display
      Incr bytMux
   Else
      bytDigit = bytDigit + 1                     'Next digit to display
      If bytDigit > bytNumDigits Then
         bytDigit = 1
      End If
      bytMux = 0                                  'Reset MUX counter
      bitShow = 1                                 'Display current digit
   End If

Return

'==============================================================================
Timer1_isr:
   'Interrupt Service Routine for Timer1 (Counter)
   'increment counter high word (Count+65536)

   Incr wrdCountHi

Return

'==============================================================================
Sub ShowFrequency
'Show frequency in 7 segment display
'Input   dwdCountSave   Frequency in units 100 Hz
'        bytDigt        Digit to be displayed in multiplex mode

   strDis = str(dwdCountSave)
   strDis = Format(strDis , "      ")             'Skip leading zeros

   bytNumber = bytDis(bytDigit)                   'single ASCII char. (overlay)
   Call ShowDigit                                 'Show current digit
   Waitms 3                                       'Adjust for display brightness
   Call ClearDigit                                'Clear current digit

End Sub

'==============================================================================
Sub ShowDigit
'Display a single digit. Is a bit messy, no complete port available
'Common anode 7 segment displays
'Anodes are switched by PNP's (neg. signal to activate anodes)
'7 segments (kathodes) switched by a ULN2803 (pos. signal to activate kathodes)
'Input:
'bytNumber  Number to be displayed
'bytDigit   Digit to be displayed


'  +------+
'  |  a   |
'  |f   b |
'  |  g   |
'  |e   c |
'  |  d   |
'  +------+
'        ---- segments -----
'Number  g  f  e  d  c  b  a
'0       0  1  1  1  1  1  1
'1       0  0  0  0  1  1  0
'2       1  0  1  1  0  1  1
'3       1  0  0  1  1  1  1
'4       1  1  0  0  1  1  0
'5       1  1  0  1  1  0  1
'6       1  1  1  1  1  0  1
'7       0  0  0  0  1  1  1
'8       1  1  1  1  1  1  1
'9       1  1  0  1  1  1  1

   Seg_a = 0                                      'All segments off
   Seg_b = 0
   Seg_c = 0
   Seg_d = 0
   Seg_e = 0
   Seg_f = 0
   Seg_g = 0
   Select Case bytNumber                          'Set segments
      Case 48                                     'ASCII 0
         Seg_a = 1
         Seg_b = 1
         Seg_c = 1
         Seg_d = 1
         Seg_e = 1
         Seg_f = 1
      Case 49                                     'ASCII 1
         Seg_b = 1
         Seg_c = 1
      Case 50                                     'ASCII 2
         Seg_a = 1
         Seg_b = 1
         Seg_d = 1
         Seg_e = 1
         Seg_g = 1
      Case 51                                     'ASCII 3
         Seg_a = 1
         Seg_b = 1
         Seg_c = 1
         Seg_d = 1
         Seg_g = 1
      Case 52                                     'ASCII 4
         Seg_b = 1
         Seg_c = 1
         Seg_f = 1
         Seg_g = 1
      Case 53                                     'ASCII 5
         Seg_a = 1
         Seg_c = 1
         Seg_d = 1
         Seg_f = 1
         Seg_g = 1
      Case 54                                     'ASCII 6
         Seg_a = 1
         Seg_c = 1
         Seg_d = 1
         Seg_e = 1
         Seg_f = 1
         Seg_g = 1
      Case 55                                     'ASCII 7
         Seg_a = 1
         Seg_b = 1
         Seg_c = 1
      Case 56                                     'ASCII 8
         Seg_a = 1
         Seg_b = 1
         Seg_c = 1
         Seg_d = 1
         Seg_e = 1
         Seg_f = 1
         Seg_g = 1
      Case 57                                     'ASCII 9
         Seg_a = 1
         Seg_b = 1
         Seg_c = 1
         Seg_d = 1
         Seg_f = 1
         Seg_g = 1
   End Select

   'Turn on digit bytDigit
   Seg_DP = 0
   Select Case bytDigit                           'Set the digit
      Case 1
         Anode_1 = 0                              'Digit 1 = on
      Case 2
         Anode_2 = 0                              'Digit 2 = on
         Seg_DP = 1                               'Decimal point on
      Case 3
         Anode_3 = 0                              'Digit 3 = on
      Case 4
         Anode_4 = 0                              'Digit 4 = on
      Case 5
         Anode_5 = 0                              'Digit 5 = on
      Case 6
         Anode_6 = 0                              'Digit 6 = on
   End Select


End Sub

'==============================================================================
Sub ClearDigit
'Clear digit bytDig

   Select Case bytDigit                           'Set the digit
      Case 1
         Anode_1 = 1                              'Digit 1 = off
      Case 2
         Anode_2 = 1                              'Digit 2 = off
      Case 3
         Anode_3 = 1                              'Digit 3 = off
      Case 4
         Anode_4 = 1                              'Digit 4 = off
      Case 5
         Anode_5 = 1                              'Digit 5 = off
      Case 6
         Anode_6 = 1                              'Digit 6 = off
   End Select
End Sub

'==============================================================================
Sub SelfTest
'Test 7 segment display

   For bytNumber = 57 To 48 Step -1               'Loop numbers 9 to 0
      For bytTmp0 = 1 to 50
         For bytDigit = 1 to bytNumDigits         'Loop the digits
            Call ShowDigit                        'Show digit
            Waitms 2
            Call ClearDigit
         Next bytTmp1
      Next bytTmp0
      Waitms 200
   Next bytNumber

End Sub

'==============================================================================

End